use std::{
    cmp::Ordering,
    fs,
    path::{Path, PathBuf},
};

use chrono::{DateTime, NaiveDateTime, Utc};
use sqlx::{prelude::FromRow, sqlite::SqliteRow, QueryBuilder};
use strum_macros::{EnumDiscriminants, EnumIter, EnumMessage};

pub trait PdfSheet {
    fn get_pdf(&self) -> &Pdf;
}

#[derive(PartialEq, Eq, PartialOrd)]
pub struct EnumSheet {
    pub id: i64,
    pub last_opened: I64DateTime,
    pub kind: SheetKind,
}

#[derive(Debug, EnumDiscriminants)]
#[strum_discriminants(derive(EnumIter, EnumMessage))]
pub enum SheetKind {
    #[strum_discriminants(strum(message = "sheets"))] // Message is the sqlite table name
    Sheet {
        pdf: Pdf,
        name: String,
        composer_id: i64,
    },
    #[strum_discriminants(strum(message = "orphans"))] // Message is the sqlite table name
    Orphan { pdf: Pdf },
    #[strum_discriminants(strum(message = "books"))] // Message is the sqlite table name
    Book {
        pdf: Pdf,
        name: String,
        composer_id: i64,
        sheet_ids: Vec<i64>,
    },
    #[strum_discriminants(strum(message = "booksheets"))] // Message is the sqlite table name
    BookSheet {
        book_id: i64,
        first_page: i64,
        last_page: i64,
    },
}

pub enum SheetKindTable {
    Sheet,
}

pub trait AnySheet: Ord {}

#[derive(sqlx::FromRow, PartialEq, Eq)]
pub struct Sheet {
    id: i64,
    #[sqlx(flatten)]
    pdf: Pdf,
    #[sqlx(try_from = "i64")]
    last_opened: I64DateTime,
    name: String,
    composer_id: i64,
}

#[derive(sqlx::FromRow, Debug, Clone, PartialEq, Eq)]
pub struct Orphan {
    id: i64,
    #[sqlx(flatten)]
    pdf: Pdf,
    #[sqlx(try_from = "i64")]
    last_opened: I64DateTime,
}

#[derive(sqlx::FromRow, Debug, Clone, PartialEq, Eq)]
pub struct Book {
    id: i64,
    #[sqlx(flatten)]
    pdf: Pdf,
    #[sqlx(try_from = "i64")]
    last_opened: I64DateTime,
    name: String,
    composer_id: i64,
    sheet_ids: Vec<i64>,
}

#[derive(sqlx::FromRow, Debug, Clone, PartialEq, Eq)]
pub struct BookSheet {
    id: i64,
    #[sqlx(try_from = "i64")]
    last_opened: I64DateTime,
    book_id: i64,
    first_page: i64,
    last_page: i64,
}

#[derive(sqlx::FromRow, Debug, Clone, PartialEq, Eq)]
pub struct Pdf {
    #[sqlx(try_from = "String")]
    path: PathBuf,
    file_size: u64,
    file_hash: String,
}

#[derive(sqlx::FromRow)]
pub struct Composer {
    id: i64,
    name: String,
}

#[derive(Debug, Clone, PartialEq, Eq)]
pub struct I64DateTime(DateTime<Utc>);

impl TryFrom<i64> for I64DateTime {
    type Error = String;

    fn try_from(value: i64) -> Result<Self, Self::Error> {
        Ok(I64DateTime(
            NaiveDateTime::from_timestamp_opt(value, 0)
                .ok_or("Failed converting i64 to DateTime")?
                .and_utc(),
        ))
    }
}

impl From<I64DateTime> for i64 {
    fn from(value: I64DateTime) -> Self {
        value.0.timestamp()
    }
}

impl Pdf {
    pub fn validate_path(&self, path: impl AsRef<Path>) -> std::io::Result<bool> {
        // First compare file size since it is faster than hashing
        let file_size = fs::metadata(path.as_ref())?.len();
        if file_size == self.file_size {
            let file_content = fs::read(path.as_ref())?;
            let file_hash = blake3::hash(&file_content);
            if file_hash.to_string() == self.file_hash {
                return Ok(true);
            }
        }

        Ok(false)
    }
}

impl TryFrom<PathBuf> for Pdf {
    type Error = std::io::Error;

    fn try_from(path: PathBuf) -> Result<Self, Self::Error> {
        let file_size = fs::metadata(path.as_path())?.len();
        let file_content = fs::read(path.as_path())?;
        let file_hash = blake3::hash(&file_content).to_string();

        Ok(Pdf {
            path,
            file_size,
            file_hash,
        })
    }
}

impl PdfSheet for Sheet {
    fn get_pdf(&self) -> &Pdf {
        &self.pdf
    }
}

impl PdfSheet for Orphan {
    fn get_pdf(&self) -> &Pdf {
        &self.pdf
    }
}

impl PdfSheet for Book {
    fn get_pdf(&self) -> &Pdf {
        &self.pdf
    }
}

impl AnySheet for Sheet {}

impl Ord for EnumSheet {
    fn cmp(&self, other: &Self) -> Ordering {
        self.last_opened.cmp(other.last_opened)
    }
}

impl EnumSheet {
    pub fn update_path_in_database_query(&self) -> String {
        todo!()
        // sqlx::query("UPDATE sheets SET path = $1 WHERE id = $2")
        //     .bind(sheet.path.to_str().unwrap().to_string())
        //     .bind(sheet.id)
    }
    pub fn update_last_opened_in_database_query(&self) -> String {
        // sqlx::query("UPDATE sheets SET last_opened = $1 WHERE id = $2")
        //     .bind(sheet.last_opened.timestamp())
        //     .bind(sheet.id)
        todo!()
    }
    pub fn insert_to_database_query(&self) -> String {
        todo!()
        // sqlx::query(
        //     "
        //         INSERT INTO orphan_files (path, file_size, file_hash, last_opened)
        //         VALUES ($1, $2, $3, $4)
        //     ",
        // )
        // .bind(file.path.to_str().unwrap().to_string())
        // .bind(file.file_size as i32)
        // .bind(file.file_hash.clone())
        // .bind(file.last_opened.timestamp())
    }

    pub fn get_database_table_name(&self) -> &str {
        todo!()
    }
}

// impl PartialOrd for EnumSheet {
//     fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
//         Some(self.cmp(other))
//     }
// }
