use chrono::NaiveDateTime;
use sqlx::{sqlite::SqliteRow, SqlitePool};

use crate::sheet::{EnumSheet, Pdf, SheetKind, SheetKindDiscriminants};

pub async fn insert_sheet(connection: SqlitePool, sheet: &EnumSheet) -> sqlx::Result<()> {
    let table = sheet.kind.into::<SheetKindDiscriminants>().get_message();
    sqlx::query(&format!(
        "
            INSERT INTO {} (path, file_size, file_hash, last_opened)
            VALUES ($1, $2, $3, $4)
        ",
        table,
    ))
    .bind(sheet.pdf.path.to_str().unwrap().to_string())
    .bind(sheet.pdf.file_size as i32)
    .bind(sheet.pdf.file_hash.clone())
    .bind(sheet.last_opened.timestamp())
    .execute(&mut connection)
    .await
    .map(|_| ())
}

pub async fn update_sheet_path(connection: SqlitePool, sheet: &EnumSheet) -> sqlx::Result<()> {
    let table = sheet.kind.into::<SheetKindDiscriminants>().get_message();
    sqlx::query(&format!("UPDATE {} SET path = $1 WHERE id = $2", table))
        .bind(sheet.kind.pdf.path.to_str().unwrap().to_string())
        .bind(sheet.id)
        .execute(&mut connection)
        .await
        .map(|_| ())
    // TODO: check for success
}

pub async fn update_sheet_last_opened(
    connection: SqlitePool,
    sheet: &EnumSheet,
) -> sqlx::Result<()> {
    let table = sheet.kind.into::<SheetKindDiscriminants>().get_message();
    sqlx::query(&format!(
        "UPDATE {} SET last_opened = $1 WHERE id = $2",
        table
    ))
    .bind(sheet.last_opened.timestamp())
    .bind(sheet.id)
    .execute(&mut connection)
    .await
    .map(|_| ())
    // TODO: check for success
}

pub async fn fetch_all_sheets(&connection: SqlitePool) -> sqlx::Result<Vec<EnumSheet>> {
    let mut sheets: Vec<EnumSheet> = Vec::new();

    for kind in SheetKindDiscriminants::iter() {
        let table = kind.get_message();

        let mut sheets_of_kind = sqlx::query(&format!("SELECT * FROM {}", table))
            .map(|row: SqliteRow| EnumSheet {
                id: row.try_get("id")?,
                last_opened: NaiveDateTime::from_timestamp_opt(
                    row.try_get::<i64, _>("last_opened")?,
                    0,
                )
                .unwrap()
                .and_utc(),
                kind: parse_kind_from_row(kind, row),
            })
            .fetch_all(&mut connection)
            .await?;

        sheets.append(&mut sheets_of_kind);
    }

    Ok(sheets)
}

fn parse_kind_from_row(kind: SheetKindDiscriminants, row: SqliteRow) -> sqlx::Result<SheetKind> {
    Ok(match kind {
        SheetKindDiscriminants::Sheet => SheetKind::Sheet {
            name: row.try_get("name")?,
            composer_id: row.try_get("composer_id")?,
            pdf: Pdf::from_row(row)?,
        },
        SheetKindDiscriminants::Orphan => SheetKind::Orphan {
            pdf: Pdf::from_row(row)?,
        },
        SheetKindDiscriminants::Book => SheetKind::Book {
            name: row.try_get("name")?,
            composer_id: row.try_get("composer_id")?,
            pdf: Pdf::from_row(row)?,
            sheet_ids: todo!(),
        },
        SheetKindDiscriminants::BookSheet => SheetKind::BookSheet {
            book_id: row.try_get("book_id")?,
            first_page: row.try_get("first_page")?,
            last_page: row.try_get("last_page")?,
        },
    })
}
